@@ -1,5 +1,5 @@
-use crate::check::{check_message, inappropriate_message};
-use crate::conn::{CommonState, ConnectionRandoms, State};
+use crate::check::{inappropriate_handshake_message, inappropriate_message};
+use crate::conn::{CommonState, ConnectionRandoms, Side, State};
 use crate::error::Error;
 use crate::hash_hs::HandshakeHash;
 #[cfg(feature = "logging")]
@@ -14,6 +14,7 @@ use crate::msgs::handshake::{DigitallySignedStruct, ServerECDHParams};
 use crate::msgs::handshake::{HandshakeMessagePayload, HandshakePayload, NewSessionTicketPayload};
 use crate::msgs::message::{Message, MessagePayload};
 use crate::msgs::persist;
+use crate::sign::Signer;
 use crate::suites::SupportedCipherSuite;
 use crate::ticketer::TimeBase;
 use crate::tls12::{self, ConnectionSecrets, Tls12CipherSuite};
@@ -25,6 +26,7 @@ use crate::client::common::ClientAuthDetails;
 use crate::client::common::ServerCertDetails;
 use crate::client::{hs, ClientConfig, ServerName};
 
+use ring::agreement::PublicKey;
 use ring::constant_time;
 
 use std::sync::Arc;
@@ -60,11 +62,10 @@ mod server_hello {
                 .write_slice(&mut self.randoms.server);
 
             // Look for TLS1.3 downgrade signal in server random
-            if tls13_supported
-                && self
-                    .randoms
-                    .has_tls12_downgrade_marker()
-            {
+            // both the server random and TLS12_DOWNGRADE_SENTINEL are
+            // public values and don't require constant time comparison
+            let has_downgrade_marker = self.randoms.server[24..] == tls12::DOWNGRADE_SENTINEL;
+            if tls13_supported && has_downgrade_marker {
                 return Err(cx
                     .common
                     .illegal_param("downgrade to TLS1.2 when TLS1.3 is supported"));
@@ -125,14 +126,14 @@ mod server_hello {
                     }
 
                     let secrets =
-                        ConnectionSecrets::new_resume(&self.randoms, suite, resuming.secret());
+                        ConnectionSecrets::new_resume(self.randoms, suite, resuming.secret());
                     self.config.key_log.log(
                         "CLIENT_RANDOM",
                         &secrets.randoms.client,
                         &secrets.master_secret,
                     );
                     cx.common
-                        .start_encryption_tls12(&secrets);
+                        .start_encryption_tls12(&secrets, Side::Client);
 
                     // Since we're resuming, we verified the certificate and
                     // proof of possession in the prior session.
@@ -265,17 +266,11 @@ struct ExpectCertificateStatusOrServerKx {
 
 impl State<ClientConnectionData> for ExpectCertificateStatusOrServerKx {
     fn handle(self: Box<Self>, cx: &mut ClientContext<'_>, m: Message) -> hs::NextStateOrError {
-        check_message(
-            &m,
-            &[ContentType::Handshake],
-            &[
-                HandshakeType::ServerKeyExchange,
-                HandshakeType::CertificateStatus,
-            ],
-        )?;
-
-        if m.is_handshake_type(HandshakeType::ServerKeyExchange) {
-            Box::new(ExpectServerKx {
+        match m.payload {
+            MessagePayload::Handshake(HandshakeMessagePayload {
+                payload: HandshakePayload::ServerKeyExchange(..),
+                ..
+            }) => Box::new(ExpectServerKx {
                 config: self.config,
                 resuming_session: self.resuming_session,
                 session_id: self.session_id,
@@ -291,9 +286,11 @@ impl State<ClientConnectionData> for ExpectCertificateStatusOrServerKx {
                 ),
                 must_issue_new_ticket: self.must_issue_new_ticket,
             })
-            .handle(cx, m)
-        } else {
-            Box::new(ExpectCertificateStatus {
+            .handle(cx, m),
+            MessagePayload::Handshake(HandshakeMessagePayload {
+                payload: HandshakePayload::CertificateStatus(..),
+                ..
+            }) => Box::new(ExpectCertificateStatus {
                 config: self.config,
                 resuming_session: self.resuming_session,
                 session_id: self.session_id,
@@ -306,7 +303,15 @@ impl State<ClientConnectionData> for ExpectCertificateStatusOrServerKx {
                 server_cert_chain: self.server_cert_chain,
                 must_issue_new_ticket: self.must_issue_new_ticket,
             })
-            .handle(cx, m)
+            .handle(cx, m),
+            payload => Err(inappropriate_handshake_message(
+                &payload,
+                &[ContentType::Handshake],
+                &[
+                    HandshakeType::ServerKeyExchange,
+                    HandshakeType::CertificateStatus,
+                ],
+            )),
         }
     }
 }
@@ -438,13 +443,9 @@ fn emit_certificate(
     common.send_msg(cert, false);
 }
 
-fn emit_clientkx(
-    transcript: &mut HandshakeHash,
-    common: &mut CommonState,
-    kxd: &kx::KeyExchangeResult,
-) {
+fn emit_clientkx(transcript: &mut HandshakeHash, common: &mut CommonState, pubkey: &PublicKey) {
     let mut buf = Vec::new();
-    let ecpoint = PayloadU8::new(Vec::from(kxd.pubkey.as_ref()));
+    let ecpoint = PayloadU8::new(Vec::from(pubkey.as_ref()));
     ecpoint.encode(&mut buf);
     let pubkey = Payload::new(buf);
 
@@ -462,21 +463,12 @@ fn emit_clientkx(
 
 fn emit_certverify(
     transcript: &mut HandshakeHash,
-    client_auth: &mut ClientAuthDetails,
+    signer: &dyn Signer,
     common: &mut CommonState,
 ) -> Result<(), Error> {
-    let (signer, message) = match (client_auth.signer.take(), transcript.take_handshake_buf()) {
-        (Some(signer), Some(msg)) => (signer, msg),
-        (None, _) => {
-            trace!("Not sending CertificateVerify, no key");
-            transcript.abandon_client_auth();
-            return Ok(());
-        }
-        (_, None) => {
-            trace!("Not sending CertificateVerify, no transcript");
-            return Ok(());
-        }
-    };
+    let message = transcript
+        .take_handshake_buf()
+        .ok_or_else(|| Error::General("Expected transcript".to_owned()))?;
 
     let scheme = signer.scheme();
     let sig = signer.sign(&message)?;
@@ -629,37 +621,21 @@ impl State<ClientConnectionData> for ExpectCertificateRequest {
         self.transcript.add_message(&m);
         debug!("Got CertificateRequest {:?}", certreq);
 
-        let mut client_auth = ClientAuthDetails::new();
-
         // The RFC jovially describes the design here as 'somewhat complicated'
         // and 'somewhat underspecified'.  So thanks for that.
         //
         // We ignore certreq.certtypes as a result, since the information it contains
         // is entirely duplicated in certreq.sigschemes.
 
-        let canames = certreq
-            .canames
-            .iter()
-            .map(|p| p.0.as_slice())
-            .collect::<Vec<&[u8]>>();
-        let maybe_certkey = self
-            .config
-            .client_auth_cert_resolver
-            .resolve(&canames, &certreq.sigschemes);
-
-        if let Some(certkey) = maybe_certkey {
-            let maybe_signer = certkey
-                .key
-                .choose_scheme(&certreq.sigschemes);
-
-            if maybe_signer.is_some() {
-                debug!("Attempting client auth");
-                client_auth.certkey = Some(certkey);
-            }
-            client_auth.signer = maybe_signer;
-        } else {
-            debug!("Client auth requested but no cert/sigscheme available");
-        }
+        const NO_CONTEXT: Option<Vec<u8>> = None; // TLS 1.2 doesn't use a context.
+        let client_auth = ClientAuthDetails::resolve(
+            self.config
+                .client_auth_cert_resolver
+                .as_ref(),
+            Some(&certreq.canames),
+            &certreq.sigschemes,
+            NO_CONTEXT,
+        );
 
         Ok(Box::new(ExpectServerDone {
             config: self.config,
@@ -695,12 +671,21 @@ struct ExpectServerDone {
 
 impl State<ClientConnectionData> for ExpectServerDone {
     fn handle(self: Box<Self>, cx: &mut ClientContext<'_>, m: Message) -> hs::NextStateOrError {
+        match m.payload {
+            MessagePayload::Handshake(HandshakeMessagePayload {
+                payload: HandshakePayload::ServerHelloDone,
+                ..
+            }) => {}
+            payload => {
+                return Err(inappropriate_handshake_message(
+                    &payload,
+                    &[ContentType::Handshake],
+                    &[HandshakeType::ServerHelloDone],
+                ));
+            }
+        }
+
         let mut st = *self;
-        check_message(
-            &m,
-            &[ContentType::Handshake],
-            &[HandshakeType::ServerHelloDone],
-        )?;
         st.transcript.add_message(&m);
 
         cx.common.check_aligned_handshake()?;
@@ -771,12 +756,12 @@ impl State<ClientConnectionData> for ExpectServerDone {
         cx.common.peer_certificates = Some(st.server_cert.cert_chain);
 
         // 4.
-        if let Some(client_auth) = &mut st.client_auth {
-            if let Some(cert_key) = &client_auth.certkey {
-                emit_certificate(&mut st.transcript, cert_key.cert.clone(), cx.common);
-            } else {
-                emit_certificate(&mut st.transcript, Vec::new(), cx.common);
-            }
+        if let Some(client_auth) = &st.client_auth {
+            let certs = match client_auth {
+                ClientAuthDetails::Empty { .. } => Vec::new(),
+                ClientAuthDetails::Verify { certkey, .. } => certkey.cert.clone(),
+            };
+            emit_certificate(&mut st.transcript, certs, cx.common);
         }
 
         // 5a.
@@ -788,40 +773,45 @@ impl State<ClientConnectionData> for ExpectServerDone {
                     Error::PeerMisbehavedError("peer chose an unsupported group".to_string())
                 })?;
         let kx = kx::KeyExchange::start(group).ok_or(Error::FailedToGetRandomBytes)?;
-        let kxd = tls12::complete_ecdh(kx, &ecdh_params.public.0)?;
 
         // 5b.
-        emit_clientkx(&mut st.transcript, cx.common, &kxd);
+        let mut transcript = st.transcript;
+        emit_clientkx(&mut transcript, cx.common, &kx.pubkey);
         // nb. EMS handshake hash only runs up to ClientKeyExchange.
-        let handshake_hash = st.transcript.get_current_hash();
+        let ems_seed = st
+            .using_ems
+            .then(|| transcript.get_current_hash());
 
         // 5c.
-        if let Some(client_auth) = &mut st.client_auth {
-            emit_certverify(&mut st.transcript, client_auth, cx.common)?;
+        if let Some(ClientAuthDetails::Verify { signer, .. }) = &st.client_auth {
+            emit_certverify(&mut transcript, signer.as_ref(), cx.common)?;
         }
 
         // 5d.
         emit_ccs(cx.common);
 
         // 5e. Now commit secrets.
-        let secrets = if st.using_ems {
-            ConnectionSecrets::new_ems(&st.randoms, &handshake_hash, suite, &kxd.shared_secret)
-        } else {
-            ConnectionSecrets::new(&st.randoms, suite, &kxd.shared_secret)
-        };
+        let secrets = ConnectionSecrets::from_key_exchange(
+            kx,
+            &ecdh_params.public.0,
+            ems_seed,
+            st.randoms,
+            suite,
+        )?;
+
         st.config.key_log.log(
             "CLIENT_RANDOM",
             &secrets.randoms.client,
             &secrets.master_secret,
         );
         cx.common
-            .start_encryption_tls12(&secrets);
+            .start_encryption_tls12(&secrets, Side::Client);
         cx.common
             .record_layer
             .start_encrypting();
 
         // 6.
-        emit_finished(&secrets, &mut st.transcript, cx.common);
+        emit_finished(&secrets, &mut transcript, cx.common);
 
         if st.must_issue_new_ticket {
             Ok(Box::new(ExpectNewTicket {
@@ -831,7 +821,7 @@ impl State<ClientConnectionData> for ExpectServerDone {
                 session_id: st.session_id,
                 server_name: st.server_name,
                 using_ems: st.using_ems,
-                transcript: st.transcript,
+                transcript,
                 resuming: false,
                 cert_verified,
                 sig_verified,
@@ -844,7 +834,7 @@ impl State<ClientConnectionData> for ExpectServerDone {
                 session_id: st.session_id,
                 server_name: st.server_name,
                 using_ems: st.using_ems,
-                transcript: st.transcript,
+                transcript,
                 ticket: None,
                 resuming: false,
                 cert_verified,
@@ -914,7 +904,15 @@ struct ExpectCcs {
 
 impl State<ClientConnectionData> for ExpectCcs {
     fn handle(self: Box<Self>, cx: &mut ClientContext<'_>, m: Message) -> hs::NextStateOrError {
-        check_message(&m, &[ContentType::ChangeCipherSpec], &[])?;
+        match m.payload {
+            MessagePayload::ChangeCipherSpec(..) => {}
+            payload => {
+                return Err(inappropriate_message(
+                    &payload,
+                    &[ContentType::ChangeCipherSpec],
+                ));
+            }
+        }
         // CCS should not be received interleaved with fragmented handshake-level
         // message.
         cx.common.check_aligned_handshake()?;
@@ -1071,8 +1069,11 @@ impl State<ClientConnectionData> for ExpectTraffic {
             MessagePayload::ApplicationData(payload) => cx
                 .common
                 .take_received_plaintext(payload),
-            _ => {
-                return Err(inappropriate_message(&m, &[ContentType::ApplicationData]));
+            payload => {
+                return Err(inappropriate_message(
+                    &payload,
+                    &[ContentType::ApplicationData],
+                ));
             }
         }
         Ok(self)
