@@ -186,7 +186,7 @@ mod test {
     #[test]
     fn test_noserversessionstorage_drops_put() {
         let c = NoServerSessionStorage {};
-        assert_eq!(c.put(vec![0x01], vec![0x02]), false);
+        assert!(!c.put(vec![0x01], vec![0x02]));
     }
 
     #[test]
@@ -209,13 +209,13 @@ mod test {
     #[test]
     fn test_serversessionmemorycache_accepts_put() {
         let c = ServerSessionMemoryCache::new(4);
-        assert_eq!(c.put(vec![0x01], vec![0x02]), true);
+        assert!(c.put(vec![0x01], vec![0x02]));
     }
 
     #[test]
     fn test_serversessionmemorycache_persists_put() {
         let c = ServerSessionMemoryCache::new(4);
-        assert_eq!(c.put(vec![0x01], vec![0x02]), true);
+        assert!(c.put(vec![0x01], vec![0x02]));
         assert_eq!(c.get(&[0x01]), Some(vec![0x02]));
         assert_eq!(c.get(&[0x01]), Some(vec![0x02]));
     }
@@ -223,19 +223,19 @@ mod test {
     #[test]
     fn test_serversessionmemorycache_overwrites_put() {
         let c = ServerSessionMemoryCache::new(4);
-        assert_eq!(c.put(vec![0x01], vec![0x02]), true);
-        assert_eq!(c.put(vec![0x01], vec![0x04]), true);
+        assert!(c.put(vec![0x01], vec![0x02]));
+        assert!(c.put(vec![0x01], vec![0x04]));
         assert_eq!(c.get(&[0x01]), Some(vec![0x04]));
     }
 
     #[test]
     fn test_serversessionmemorycache_drops_to_maintain_size_invariant() {
         let c = ServerSessionMemoryCache::new(2);
-        assert_eq!(c.put(vec![0x01], vec![0x02]), true);
-        assert_eq!(c.put(vec![0x03], vec![0x04]), true);
-        assert_eq!(c.put(vec![0x05], vec![0x06]), true);
-        assert_eq!(c.put(vec![0x07], vec![0x08]), true);
-        assert_eq!(c.put(vec![0x09], vec![0x0a]), true);
+        assert!(c.put(vec![0x01], vec![0x02]));
+        assert!(c.put(vec![0x03], vec![0x04]));
+        assert!(c.put(vec![0x05], vec![0x06]));
+        assert!(c.put(vec![0x07], vec![0x08]));
+        assert!(c.put(vec![0x09], vec![0x0a]));
 
         let count = c.get(&[0x01]).iter().count()
             + c.get(&[0x03]).iter().count()
@@ -249,7 +249,7 @@ mod test {
     #[test]
     fn test_neverproducestickets_does_nothing() {
         let npt = NeverProducesTickets {};
-        assert_eq!(false, npt.enabled());
+        assert!(!npt.enabled());
         assert_eq!(0, npt.lifetime());
         assert_eq!(None, npt.encrypt(&[]));
         assert_eq!(None, npt.decrypt(&[]));
