@@ -1,6 +1,4 @@
-#[cfg(feature = "quic")]
-use crate::check::check_message;
-use crate::check::{inappropriate_handshake_message, inappropriate_message};
+use crate::check::inappropriate_handshake_message;
 use crate::conn::{CommonState, ConnectionRandoms, State};
 use crate::error::Error;
 use crate::hash_hs::HandshakeHash;
@@ -12,16 +10,17 @@ use crate::msgs::enums::{AlertDescription, KeyUpdateRequest};
 use crate::msgs::enums::{ContentType, HandshakeType, ProtocolVersion};
 use crate::msgs::handshake::HandshakeMessagePayload;
 use crate::msgs::handshake::HandshakePayload;
-use crate::msgs::handshake::NewSessionTicketPayloadTLS13;
+use crate::msgs::handshake::{NewSessionTicketExtension, NewSessionTicketPayloadTLS13};
 use crate::msgs::message::{Message, MessagePayload};
 use crate::msgs::persist;
 use crate::rand;
 use crate::server::ServerConfig;
+use crate::ticketer;
 use crate::tls13::key_schedule::{KeyScheduleTraffic, KeyScheduleTrafficWithClientFinishedPending};
 use crate::tls13::Tls13CipherSuite;
 use crate::verify;
 #[cfg(feature = "quic")]
-use crate::{conn::Protocol, msgs::handshake::NewSessionTicketExtension};
+use crate::{check::inappropriate_message, conn::Protocol};
 
 use super::hs::{self, HandshakeHashOrBuffer, ServerContext};
 use super::server_conn::ServerConnectionData;
@@ -58,11 +57,18 @@ mod client_hello {
     use crate::server::common::ActiveCertifiedKey;
     use crate::sign;
     use crate::tls13::key_schedule::{
-        KeyScheduleEarly, KeyScheduleHandshake, KeyScheduleNonSecret,
+        KeyScheduleEarly, KeyScheduleHandshake, KeySchedulePreHandshake,
     };
 
     use super::*;
 
+    #[derive(PartialEq)]
+    pub(super) enum EarlyDataDecision {
+        Disabled,
+        RequestedButRejected,
+        Accepted,
+    }
+
     pub(in crate::server) struct CompleteClientHelloHandling {
         pub(in crate::server) config: Arc<ServerConfig>,
         pub(in crate::server) transcript: HandshakeHash,
@@ -73,6 +79,21 @@ mod client_hello {
         pub(in crate::server) extra_exts: Vec<ServerExtension>,
     }
 
+    fn max_early_data_size(configured: u32) -> usize {
+        if configured != 0 {
+            configured as usize
+        } else {
+            // The relevant max_early_data_size may in fact be unknowable: if
+            // we (the server) have turned off early_data but the client has
+            // a stale ticket from when we allowed early_data: we'll naturally
+            // reject early_data but need an upper bound on the amount of data
+            // to drop.
+            //
+            // Use a single maximum-sized message.
+            16384
+        }
+    }
+
     impl CompleteClientHelloHandling {
         fn check_binder(
             &self,
@@ -130,14 +151,14 @@ mod client_hello {
 
             let groups_ext = client_hello
                 .get_namedgroups_extension()
-                .ok_or_else(|| hs::incompatible(&mut cx.common, "client didn't describe groups"))?;
+                .ok_or_else(|| hs::incompatible(cx.common, "client didn't describe groups"))?;
 
             let tls13_schemes = sign::supported_sign_tls13();
             sigschemes_ext.retain(|scheme| tls13_schemes.contains(scheme));
 
             let shares_ext = client_hello
                 .get_keyshare_extension()
-                .ok_or_else(|| hs::incompatible(&mut cx.common, "client didn't send keyshares"))?;
+                .ok_or_else(|| hs::incompatible(cx.common, "client didn't send keyshares"))?;
 
             if client_hello.has_keyshare_extension_with_duplicates() {
                 return Err(cx
@@ -145,6 +166,15 @@ mod client_hello {
                     .illegal_param("client sent duplicate keyshares"));
             }
 
+            let early_data_requested = client_hello.early_data_extension_offered();
+
+            // EarlyData extension is illegal in second ClientHello
+            if self.done_retry && early_data_requested {
+                return Err(cx
+                    .common
+                    .illegal_param("client sent EarlyData in second ClientHello"));
+            }
+
             // choose a share that we support
             let chosen_share = self
                 .config
@@ -180,11 +210,14 @@ mod client_hello {
                         emit_hello_retry_request(
                             &mut self.transcript,
                             self.suite,
-                            &mut cx.common,
+                            cx.common,
                             group.name,
                         );
-                        emit_fake_ccs(&mut cx.common);
-                        return Ok(Box::new(hs::ExpectClientHello {
+                        emit_fake_ccs(cx.common);
+
+                        let skip_early_data = max_early_data_size(self.config.max_early_data_size);
+
+                        let next = Box::new(hs::ExpectClientHello {
                             config: self.config,
                             transcript: HandshakeHashOrBuffer::Hash(self.transcript),
                             #[cfg(feature = "tls12")]
@@ -194,11 +227,20 @@ mod client_hello {
                             done_retry: true,
                             send_ticket: self.send_ticket,
                             extra_exts: self.extra_exts,
-                        }));
+                        });
+
+                        return if early_data_requested {
+                            Ok(Box::new(ExpectAndSkipRejectedEarlyData {
+                                skip_data_left: skip_early_data,
+                                next,
+                            }))
+                        } else {
+                            Ok(next)
+                        };
                     }
 
                     return Err(hs::incompatible(
-                        &mut cx.common,
+                        cx.common,
                         "no kx group overlap with client",
                     ));
                 }
@@ -206,6 +248,8 @@ mod client_hello {
 
             let mut chosen_psk_index = None;
             let mut resumedata = None;
+            let time_now = ticketer::TimeBase::now()?;
+
             if let Some(psk_offer) = client_hello.get_psk() {
                 if !client_hello.check_psk_ext_is_last() {
                     return Err(cx
@@ -214,10 +258,7 @@ mod client_hello {
                 }
 
                 if psk_offer.binders.is_empty() {
-                    return Err(hs::decode_error(
-                        &mut cx.common,
-                        "psk extension missing binder",
-                    ));
+                    return Err(hs::decode_error(cx.common, "psk extension missing binder"));
                 }
 
                 if psk_offer.binders.len() != psk_offer.identities.len() {
@@ -229,6 +270,9 @@ mod client_hello {
                 for (i, psk_id) in psk_offer.identities.iter().enumerate() {
                     let resume = match self
                         .attempt_tls13_ticket_decryption(&psk_id.identity.0)
+                        .map(|resumedata| {
+                            resumedata.set_freshness(psk_id.obfuscated_ticket_age, time_now)
+                        })
                         .filter(|resumedata| {
                             hs::can_resume(self.suite.into(), &cx.data.sni, false, resumedata)
                         }) {
@@ -285,12 +329,12 @@ mod client_hello {
                 &self.config,
             )?;
             if !self.done_retry {
-                emit_fake_ccs(&mut cx.common);
+                emit_fake_ccs(cx.common);
             }
 
             let (mut ocsp_response, mut sct_list) =
                 (server_key.get_ocsp(), server_key.get_sct_list());
-            emit_encrypted_extensions(
+            let doing_early_data = emit_encrypted_extensions(
                 &mut self.transcript,
                 self.suite,
                 cx,
@@ -307,14 +351,14 @@ mod client_hello {
                     emit_certificate_req_tls13(&mut self.transcript, cx, &self.config)?;
                 emit_certificate_tls13(
                     &mut self.transcript,
-                    &mut cx.common,
+                    cx.common,
                     server_key.get_cert(),
                     ocsp_response,
                     sct_list,
                 );
                 emit_certificate_verify_tls13(
                     &mut self.transcript,
-                    &mut cx.common,
+                    cx.common,
                     server_key.get_key(),
                     &sigschemes_ext,
                 )?;
@@ -323,6 +367,36 @@ mod client_hello {
                 false
             };
 
+            // If we're not doing early data, then the next messages we receive
+            // are encrypted with the handshake keys.
+            match doing_early_data {
+                EarlyDataDecision::Disabled => {
+                    cx.common
+                        .record_layer
+                        .set_message_decrypter(
+                            self.suite
+                                .derive_decrypter(key_schedule.client_key()),
+                        );
+                    cx.data.early_data.reject();
+                }
+                EarlyDataDecision::RequestedButRejected => {
+                    debug!("Client requested early_data, but not accepted: switching to handshake keys with trial decryption");
+                    cx.common
+                        .record_layer
+                        .set_message_decrypter_with_trial_decryption(
+                            self.suite
+                                .derive_decrypter(key_schedule.client_key()),
+                            max_early_data_size(self.config.max_early_data_size),
+                        );
+                    cx.data.early_data.reject();
+                }
+                EarlyDataDecision::Accepted => {
+                    cx.data
+                        .early_data
+                        .accept(self.config.max_early_data_size as usize);
+                }
+            }
+
             cx.common.check_aligned_handshake()?;
             let key_schedule_traffic = emit_finished_tls13(
                 &mut self.transcript,
@@ -333,6 +407,13 @@ mod client_hello {
                 &self.config,
             );
 
+            if !doing_client_auth && self.config.send_half_rtt_data {
+                // Application data can be sent immediately after Finished, in one
+                // flight.  However, if client auth is enabled, we don't want to send
+                // application data to an unauthenticated peer.
+                cx.common.start_outgoing_traffic();
+            }
+
             if doing_client_auth {
                 Ok(Box::new(ExpectCertificate {
                     config: self.config,
@@ -341,6 +422,17 @@ mod client_hello {
                     key_schedule: key_schedule_traffic,
                     send_ticket: self.send_ticket,
                 }))
+            } else if doing_early_data == EarlyDataDecision::Accepted && !cx.common.is_quic() {
+                // Not used for QUIC: RFC 9001 §8.3: Clients MUST NOT send the EndOfEarlyData
+                // message. A server MUST treat receipt of a CRYPTO frame in a 0-RTT packet as a
+                // connection error of type PROTOCOL_VIOLATION.
+                Ok(Box::new(ExpectEarlyData {
+                    config: self.config,
+                    transcript: self.transcript,
+                    suite: self.suite,
+                    key_schedule: key_schedule_traffic,
+                    send_ticket: self.send_ticket,
+                }))
             } else {
                 Ok(Box::new(ExpectFinished {
                     config: self.config,
@@ -366,14 +458,12 @@ mod client_hello {
     ) -> Result<KeyScheduleHandshake, Error> {
         let mut extensions = Vec::new();
 
-        // Do key exchange
-        let kxr = kx::KeyExchange::choose(share.group, &config.kx_groups)
+        // Prepare key exchange
+        let kx = kx::KeyExchange::choose(share.group, &config.kx_groups)
             .and_then(kx::KeyExchange::start)
-            .ok_or(Error::FailedToGetRandomBytes)?
-            .complete(&share.payload.0)
-            .ok_or_else(|| Error::PeerMisbehavedError("key exchange failed".to_string()))?;
+            .ok_or(Error::FailedToGetRandomBytes)?;
 
-        let kse = KeyShareEntry::new(share.group, kxr.pubkey.as_ref());
+        let kse = KeyShareEntry::new(share.group, kx.pubkey.as_ref());
         extensions.push(ServerExtension::KeyShare(kse));
         extensions.push(ServerExtension::SupportedVersions(ProtocolVersion::TLSv1_3));
 
@@ -398,7 +488,6 @@ mod client_hello {
 
         cx.common.check_aligned_handshake()?;
 
-        #[cfg(feature = "quic")]
         let client_hello_hash = transcript.get_hash_given(&[]);
 
         trace!("sending server hello {:?}", sh);
@@ -406,48 +495,54 @@ mod client_hello {
         cx.common.send_msg(sh, false);
 
         // Start key schedule
-        let key_schedule = if let Some(psk) = resuming_psk {
+        let (key_schedule_pre_handshake, early_data_client_key) = if let Some(psk) = resuming_psk {
             let early_key_schedule = KeyScheduleEarly::new(suite.hkdf_algorithm, psk);
+            let client_early_traffic_secret = early_key_schedule.client_early_traffic_secret(
+                &client_hello_hash,
+                &*config.key_log,
+                &randoms.client,
+            );
 
-            #[cfg(feature = "quic")]
-            {
-                if cx.common.protocol == Protocol::Quic {
-                    let client_early_traffic_secret = early_key_schedule
-                        .client_early_traffic_secret(
-                            &client_hello_hash,
-                            &*config.key_log,
-                            &randoms.client,
-                        );
-                    // If 0-RTT should be rejected, this will be clobbered by ExtensionProcessing
-                    // before the application can see.
-                    cx.common.quic.early_secret = Some(client_early_traffic_secret);
-                }
-            }
-
-            early_key_schedule.into_handshake(&kxr.shared_secret)
+            (
+                KeySchedulePreHandshake::from(early_key_schedule),
+                Some(client_early_traffic_secret),
+            )
         } else {
-            KeyScheduleNonSecret::new(suite.hkdf_algorithm).into_handshake(&kxr.shared_secret)
+            (KeySchedulePreHandshake::new(suite.hkdf_algorithm), None)
         };
 
+        // Do key exchange
+        let key_schedule = kx.complete(&share.payload.0, |secret| {
+            Ok(key_schedule_pre_handshake.into_handshake(secret))
+        })?;
+
         let handshake_hash = transcript.get_current_hash();
-        let (key_schedule, client_key, server_key) = key_schedule.derive_handshake_secrets(
+        let (key_schedule, _client_key, server_key) = key_schedule.derive_handshake_secrets(
             handshake_hash,
             &*config.key_log,
             &randoms.client,
         );
 
-        // Encrypt with our own key, decrypt with the peer's key
+        // Set up to encrypt with handshake secrets, but decrypt with early_data keys.
+        // If not doing early_data after all, this is corrected later to the handshake
+        // keys (now stored in key_schedule).
         cx.common
             .record_layer
             .set_message_encrypter(suite.derive_encrypter(&server_key));
-        cx.common
-            .record_layer
-            .set_message_decrypter(suite.derive_decrypter(&client_key));
+
+        if let Some(key) = &early_data_client_key {
+            cx.common
+                .record_layer
+                .set_message_decrypter(suite.derive_decrypter(key));
+        }
 
         #[cfg(feature = "quic")]
-        {
+        if cx.common.is_quic() {
+            // If 0-RTT should be rejected, this will be clobbered by ExtensionProcessing
+            // before the application can see.
+            cx.common.quic.early_secret = early_data_client_key;
             cx.common.quic.hs_secrets =
-                Some(quic::Secrets::new(client_key, server_key, suite, false));
+                Some(quic::Secrets::new(_client_key, server_key, suite, false));
         }
 
         Ok(key_schedule)
@@ -498,6 +593,66 @@ mod client_hello {
         common.send_msg(m, false);
     }
 
+    fn decide_if_early_data_allowed(
+        cx: &mut ServerContext<'_>,
+        client_hello: &ClientHelloPayload,
+        resumedata: Option<&persist::ServerSessionValue>,
+        suite: &'static Tls13CipherSuite,
+        config: &ServerConfig,
+    ) -> EarlyDataDecision {
+        let early_data_requested = client_hello.early_data_extension_offered();
+        let rejected_or_disabled = match early_data_requested {
+            true => EarlyDataDecision::RequestedButRejected,
+            false => EarlyDataDecision::Disabled,
+        };
+
+        let resume = match resumedata {
+            Some(resume) => resume,
+            None => {
+                // never any early data if not resuming.
+                return rejected_or_disabled;
+            }
+        };
+
+        /* Non-zero max_early_data_size controls whether early_data is allowed at all.
+         * We also require stateful resumption. */
+        let early_data_configured = config.max_early_data_size > 0 && !config.ticketer.enabled();
+
+        /* "For PSKs provisioned via NewSessionTicket, a server MUST validate
+         *  that the ticket age for the selected PSK identity (computed by
+         *  subtracting ticket_age_add from PskIdentity.obfuscated_ticket_age
+         *  modulo 2^32) is within a small tolerance of the time since the ticket
+         *  was issued (see Section 8)." -- this is implemented in ServerSessionValue::set_freshness()
+         *  and related.
+         *
+         * "In order to accept early data, the server [...] MUST verify that the
+         *  following values are the same as those associated with the
+         *  selected PSK:
+         *
+         *  - The TLS version number
+         *  - The selected cipher suite
+         *  - The selected ALPN [RFC7301] protocol, if any"
+         *
+         * (RFC8446, 4.2.10) */
+        let early_data_possible = early_data_requested
+            && resume.is_fresh()
+            && Some(resume.version) == cx.common.negotiated_version
+            && resume.cipher_suite == suite.common.suite
+            && resume.alpn.as_ref().map(|x| &x.0) == cx.common.alpn_protocol.as_ref();
+
+        if early_data_configured && early_data_possible && !cx.data.early_data.was_rejected() {
+            EarlyDataDecision::Accepted
+        } else {
+            #[cfg(feature = "quic")]
+            if cx.common.is_quic() {
+                // Clobber value set in tls13::emit_server_hello
+                cx.common.quic.early_secret = None;
+            }
+
+            rejected_or_disabled
+        }
+    }
+
     fn emit_encrypted_extensions(
         transcript: &mut HandshakeHash,
         suite: &'static Tls13CipherSuite,
@@ -508,12 +663,11 @@ mod client_hello {
         resumedata: Option<&persist::ServerSessionValue>,
         extra_exts: Vec<ServerExtension>,
         config: &ServerConfig,
-    ) -> Result<(), Error> {
+    ) -> Result<EarlyDataDecision, Error> {
         let mut ep = hs::ExtensionProcessing::new();
         ep.process_common(
             config,
             cx,
-            suite.into(),
             ocsp_response,
             sct_list,
             hello,
@@ -521,6 +675,11 @@ mod client_hello {
             extra_exts,
         )?;
 
+        let early_data = decide_if_early_data_allowed(cx, hello, resumedata, suite, config);
+        if early_data == EarlyDataDecision::Accepted {
+            ep.exts.push(ServerExtension::EarlyData);
+        }
+
         let ee = Message {
             version: ProtocolVersion::TLSv1_3,
             payload: MessagePayload::Handshake(HandshakeMessagePayload {
@@ -532,7 +691,7 @@ mod client_hello {
         trace!("sending encrypted extensions {:?}", ee);
         transcript.add_message(&ee);
         cx.common.send_msg(ee, true);
-        Ok(())
+        Ok(early_data)
     }
 
     fn emit_certificate_req_tls13(
@@ -711,6 +870,28 @@ mod client_hello {
     }
 }
 
+struct ExpectAndSkipRejectedEarlyData {
+    skip_data_left: usize,
+    next: Box<hs::ExpectClientHello>,
+}
+
+impl State<ServerConnectionData> for ExpectAndSkipRejectedEarlyData {
+    fn handle(mut self: Box<Self>, cx: &mut ServerContext<'_>, m: Message) -> hs::NextStateOrError {
+        /* "The server then ignores early data by skipping all records with an external
+         *  content type of "application_data" (indicating that they are encrypted),
+         *  up to the configured max_early_data_size."
+         * (RFC8446, 14.2.10) */
+        if let MessagePayload::ApplicationData(ref skip_data) = m.payload {
+            if skip_data.0.len() <= self.skip_data_left {
+                self.skip_data_left -= skip_data.0.len();
+                return Ok(self);
+            }
+        }
+
+        self.next.handle(cx, m)
+    }
+}
+
 struct ExpectCertificate {
     config: Arc<ServerConfig>,
     transcript: HandshakeHash,
@@ -775,7 +956,7 @@ impl State<ServerConnectionData> for ExpectCertificate {
             .verifier
             .verify_client_cert(end_entity, intermediates, now)
             .map_err(|err| {
-                hs::incompatible(&mut cx.common, "certificate invalid");
+                hs::incompatible(cx.common, "certificate invalid");
                 err
             })?;
 
@@ -837,6 +1018,65 @@ impl State<ServerConnectionData> for ExpectCertificateVerify {
     }
 }
 
+// --- Process (any number of) early ApplicationData messages,
+//     followed by a terminating handshake EndOfEarlyData message ---
+
+struct ExpectEarlyData {
+    config: Arc<ServerConfig>,
+    transcript: HandshakeHash,
+    suite: &'static Tls13CipherSuite,
+    key_schedule: KeyScheduleTrafficWithClientFinishedPending,
+    send_ticket: bool,
+}
+
+impl State<ServerConnectionData> for ExpectEarlyData {
+    fn handle(mut self: Box<Self>, cx: &mut ServerContext<'_>, m: Message) -> hs::NextStateOrError {
+        match m.payload {
+            MessagePayload::ApplicationData(payload) => {
+                match cx
+                    .data
+                    .early_data
+                    .take_received_plaintext(payload)
+                {
+                    true => Ok(self),
+                    false => {
+                        cx.common
+                            .send_fatal_alert(AlertDescription::UnexpectedMessage);
+                        Err(Error::PeerMisbehavedError(
+                            "too much early_data received".into(),
+                        ))
+                    }
+                }
+            }
+            MessagePayload::Handshake(HandshakeMessagePayload {
+                typ: HandshakeType::EndOfEarlyData,
+                payload: HandshakePayload::EndOfEarlyData,
+            }) => {
+                cx.common
+                    .record_layer
+                    .set_message_decrypter(
+                        self.suite
+                            .derive_decrypter(self.key_schedule.client_key()),
+                    );
+
+                self.transcript.add_message(&m);
+                Ok(Box::new(ExpectFinished {
+                    config: self.config,
+                    suite: self.suite,
+                    key_schedule: self.key_schedule,
+                    transcript: self.transcript,
+                    send_ticket: self.send_ticket,
+                }))
+            }
+            payload => Err(inappropriate_handshake_message(
+                &payload,
+                &[ContentType::ApplicationData, ContentType::Handshake],
+                &[HandshakeType::EndOfEarlyData],
+            )),
+        }
+    }
+}
+
 // --- Process client's Finished ---
 fn get_server_session_value(
     transcript: &mut HandshakeHash,
@@ -844,6 +1084,8 @@ fn get_server_session_value(
     key_schedule: &KeyScheduleTraffic,
     cx: &ServerContext<'_>,
     nonce: &[u8],
+    time_now: ticketer::TimeBase,
+    age_obfuscation_offset: u32,
 ) -> persist::ServerSessionValue {
     let version = ProtocolVersion::TLSv1_3;
 
@@ -859,6 +1101,8 @@ fn get_server_session_value(
         cx.common.peer_certificates.clone(),
         cx.common.alpn_protocol.clone(),
         cx.data.resumption_data.clone(),
+        time_now,
+        age_obfuscation_offset,
     )
 }
 
@@ -877,10 +1121,13 @@ impl ExpectFinished {
         cx: &mut ServerContext<'_>,
         key_schedule: &KeyScheduleTraffic,
         config: &ServerConfig,
-    ) -> Result<(), rand::GetRandomFailed> {
+    ) -> Result<(), Error> {
         let nonce = rand::random_vec(32)?;
+        let now = ticketer::TimeBase::now()?;
+        let age_add = rand::random_u32()?;
         let plain =
-            get_server_session_value(transcript, suite, key_schedule, cx, &nonce).get_encoding();
+            get_server_session_value(transcript, suite, key_schedule, cx, &nonce, now, age_add)
+                .get_encoding();
 
         let stateless = config.ticketer.enabled();
         let (ticket, lifetime) = if stateless {
@@ -902,19 +1149,22 @@ impl ExpectFinished {
             (id, stateful_lifetime)
         };
 
-        let age_add = rand::random_u32()?; // nb, we don't do 0-RTT data, so whatever
-        #[allow(unused_mut)]
         let mut payload = NewSessionTicketPayloadTLS13::new(lifetime, age_add, nonce, ticket);
-        #[cfg(feature = "quic")]
-        {
-            if config.max_early_data_size > 0 && cx.common.protocol == Protocol::Quic {
+
+        if config.max_early_data_size > 0 {
+            if !stateless {
                 payload
                     .exts
                     .push(NewSessionTicketExtension::EarlyData(
                         config.max_early_data_size,
                     ));
+            } else {
+                // We implement RFC8446 section 8.1: by enforcing that 0-RTT is
+                // only possible if using stateful resumption
+                warn!("early_data with stateless resumption is not allowed");
             }
         }
+
         let m = Message {
             version: ProtocolVersion::TLSv1_3,
             payload: MessagePayload::Handshake(HandshakeMessagePayload {
@@ -970,6 +1220,7 @@ impl State<ServerConnectionData> for ExpectFinished {
             )?;
         }
 
+        // Application data may now flow, even if we have client auth enabled.
         cx.common.start_traffic();
 
         #[cfg(feature = "quic")]
@@ -1049,21 +1300,15 @@ impl State<ServerConnectionData> for ExpectTraffic {
             MessagePayload::ApplicationData(payload) => cx
                 .common
                 .take_received_plaintext(payload),
-            MessagePayload::Handshake(payload) => match payload.payload {
-                HandshakePayload::KeyUpdate(key_update) => {
-                    self.handle_key_update(cx.common, &key_update)?
-                }
-                _ => {
-                    return Err(inappropriate_handshake_message(
-                        &payload,
-                        &[HandshakeType::KeyUpdate],
-                    ));
-                }
-            },
-            _ => {
-                return Err(inappropriate_message(
-                    &m,
+            MessagePayload::Handshake(HandshakeMessagePayload {
+                payload: HandshakePayload::KeyUpdate(key_update),
+                ..
+            }) => self.handle_key_update(cx.common, &key_update)?,
+            payload => {
+                return Err(inappropriate_handshake_message(
+                    &payload,
                     &[ContentType::ApplicationData, ContentType::Handshake],
+                    &[HandshakeType::KeyUpdate],
                 ));
             }
         }
@@ -1106,8 +1351,7 @@ struct ExpectQuicTraffic {
 impl State<ServerConnectionData> for ExpectQuicTraffic {
     fn handle(self: Box<Self>, _cx: &mut ServerContext<'_>, m: Message) -> hs::NextStateOrError {
         // reject all messages
-        check_message(&m, &[], &[])?;
-        unreachable!();
+        Err(inappropriate_message(&m.payload, &[]))
     }
 
     fn export_keying_material(
