@@ -72,6 +72,31 @@ where
 
         self.conn.reader().read(buf)
     }
+
+    #[cfg(read_buf)]
+    fn read_buf(&mut self, buf: &mut std::io::ReadBuf<'_>) -> Result<()> {
+        self.complete_prior_io()?;
+
+        // We call complete_io() in a loop since a single call may read only
+        // a partial packet from the underlying transport. A full packet is
+        // needed to get more plaintext, which we must do if EOF has not been
+        // hit. Otherwise, we will prematurely signal EOF by returning without
+        // writing anything. We determine if EOF has actually been hit by
+        // checking if 0 bytes were read from the underlying transport.
+        while self.conn.wants_read() {
+            let at_eof = self.conn.complete_io(self.sock)?.0 == 0;
+            if at_eof {
+                if let Ok(io_state) = self.conn.process_new_packets() {
+                    if at_eof && io_state.plaintext_bytes_to_read() == 0 {
+                        return Ok(());
+                    }
+                }
+                break;
+            }
+        }
+
+        self.conn.reader().read_buf(buf)
+    }
 }
 
 impl<'a, C, T, S> Write for Stream<'a, C, T>
@@ -183,6 +208,11 @@ where
     fn read(&mut self, buf: &mut [u8]) -> Result<usize> {
         self.as_stream().read(buf)
     }
+
+    #[cfg(read_buf)]
+    fn read_buf(&mut self, buf: &mut std::io::ReadBuf<'_>) -> Result<()> {
+        self.as_stream().read_buf(buf)
+    }
 }
 
 impl<C, T, S> Write for StreamOwned<C, T>
